@@ -19,7 +19,6 @@ as;
 break;
 case;
 catch;
-class ClassName;
 clone;
 const;
 continue;
@@ -39,7 +38,6 @@ endswitch;
 endwhile;
 eval();
 exit();
-extends ClassName;
 final;
 finally;
 for;
@@ -48,15 +46,12 @@ function;
 global;
 goto;
 if;
-implements ClassName;
 include;
 include_once;
 instanceof ClassName;
 insteadof ClassName;
-interface ClassName;
 isset();
 list();
-namespace ClassName;
 new ClassName;
 print;
 private;
@@ -68,10 +63,8 @@ return;
 static;
 switch;
 throw;
-trait ClassName;
 try;
 unset();
-use ClassName;
 var;
 while;
 xor;
