@@ -63,22 +63,25 @@ These are the ###php-mode-test### comments. Valid magics are
 listed in `php-mode-test-valid-magics'; no other directives will
 be processed."
   (cl-letf (((symbol-function 'indent)
-             (lambda (offset) (equal (current-indentation) offset))))
+             (lambda (offset)
+               (let ((current-offset (current-indentation)))
+                 (unless (eq current-offset offset)
+                   (warn "line: %d context: %s\n" (line-number-at-pos) (c-guess-basic-syntax))
+                   (list :line (line-number-at-pos)
+                         :expected offset
+                         :actual current-offset))))))
     (let (directives answers)
       (save-excursion
         (goto-char (point-min))
-        (while (re-search-forward php-mode-test-magic-regexp nil t)
-          (setq directives (read (buffer-substring (match-beginning 1)
-                                                   (match-end 1))))
-          (setq answers
-                (append (mapcar (lambda (curr)
-                                  (let ((fn (car curr))
-                                        (args (mapcar 'eval (cdr-safe curr))))
-                                    (if (memq fn php-mode-test-valid-magics)
-                                        (apply fn args))))
-                                directives)
-                        answers))))
-      answers)))
+        (cl-loop while (re-search-forward php-mode-test-magic-regexp nil t)
+                 for directives = (read (buffer-substring (match-beginning 1) (match-end 1)))
+                 for result = (mapcar (lambda (expr)
+                                        (let ((fn (car expr))
+                                              (args (mapcar 'eval (cdr-safe expr))))
+                                          (if (memq fn php-mode-test-valid-magics)
+                                              (apply fn args))))
+                                      directives)
+                 append (cl-remove-if #'null result))))))
 
 (defun php-mode-test--buffer-face-list (buffer)
   "Return list of (STRING . FACE) from `BUFFER'."
@@ -154,8 +157,9 @@ file name and check that the faces of the fonts in the buffer match."
      ,(if indent
           '(let ((inhibit-message t)) (indent-region (point-min) (point-max))))
      ,(if magic
-          '(should (cl-reduce (lambda (l r) (and l r))
-                              (php-mode-test-process-magics))))
+          `(should (equal
+                    (cons ,file nil)
+                    (cons ,file (php-mode-test-process-magics)))))
      ,(if faces
           `(should (equal
                     (cons ,file
@@ -319,12 +323,11 @@ style from Drupal."
   "All static method should appear on imenu whether 'static' keyword is placed before or after visibility"
   (with-php-mode-test ("issue-83.php")
     (let* ((index-alist (imenu--make-index-alist))
-           (public-methods (mapcar 'car (cdr (assoc "Public Methods" index-alist))))
-           (all-methods (mapcar 'car (cdr (assoc "All Methods" index-alist)))))
-      (should (member "staticBeforeVisibility" public-methods))
-      (should (member "staticBeforeVisibility" all-methods))
-      (should (member "staticAfterVisibility" public-methods))
-      (should (member "staticAfterVisibility" all-methods)))))
+           (all-methods (mapcar 'car (cdr (assoc "Methods" index-alist)))))
+      (should (equal all-methods
+                     (list
+                      "static public function staticBeforeVisibility()"
+                      "public static function staticAfterVisibility()"))))))
 
 (ert-deftest php-mode-test-issue-99 ()
   "Proper indentation for 'foreach' statements without braces."
@@ -569,15 +572,13 @@ style from Drupal."
 
 (ert-deftest php-mode-test-issue-357 ()
   "Match version-specific interpreters."
-  (dolist (on (if (version< emacs-version "24.4")
-                  '("php" "php5" "php7")
-                '("php" "php3" "php5" "php7" "php-5" "php-5.5" "php7.0.1")))
+  (dolist (on '("php" "php3" "php5" "php7" "php-5" "php-5.5" "php7.0.1"))
     (with-temp-buffer
       (insert "#!" on)
       (set-auto-mode)
       (should (eq 'php-mode major-mode))))
 
-  (dolist (off '("php2" "xphp5" "foo" "php8" "php7x"))
+  (dolist (off '("php2" "xphp5" "foo" "php7x"))
     (with-temp-buffer
       (insert "#!" off)
       (set-auto-mode)
@@ -649,14 +650,25 @@ Meant for `php-mode-test-issue-503'."
   "Proper alignment object -> accessor."
   (with-php-mode-test ("indent/issue-623.php" :indent t :magic t)))
 
+(ert-deftest php-mode-test-issue-702 ()
+  "Proper alignment arglist."
+  (with-php-mode-test ("indent/issue-702.php" :indent t :magic t)))
+
 (ert-deftest php-mode-test-php74 ()
   "Test highlighting language constructs added in PHP 7.4."
   (with-php-mode-test ("7.4/arrow-function.php" :faces t))
   (with-php-mode-test ("7.4/typed-property.php" :faces t)))
 
+(ert-deftest php-mode-test-php80 ()
+  "Test highlighting language constructs added in PHP 8.0."
+  (with-php-mode-test ("8.0/attribute/class.php" :faces t))
+  (with-php-mode-test ("8.0/attribute/function.php" :faces t))
+  (with-php-mode-test ("8.0/attribute/function2.php" :faces t)))
+
 (ert-deftest php-mode-test-php81 ()
   "Test highlighting language constructs added in PHP 8.1."
-  (with-php-mode-test ("8.1/enum.php" :faces t)))
+  (with-php-mode-test ("8.1/enum.php" :faces t))
+  (with-php-mode-test ("8.1/readonly.php" :faces t)))
 
 (ert-deftest php-mode-test-lang ()
   "Test highlighting for language constructs."
@@ -678,4 +690,15 @@ Meant for `php-mode-test-issue-503'."
   (with-php-mode-test ("lang/errorcontrol.php" :faces t))
   (with-php-mode-test ("lang/magical-constants/echo.php" :faces t)))
 
+;; For developers: How to make .faces list file.
+;;
+;; 1. Press `M-x eval-buffer' in this file bufffer.
+;; 2. Copy follows code snippet:
+;;     (setq x (php-mode-test--buffer-face-list (current-buffer)))
+;; 3. Visit target buffer of testing PHP file.
+;; 4. Press `M-:' (or `M-x eval-expression') and yank killed the code snippet.
+;; 5. Press `M-x ielm' and input `x' and RET key.
+;; 6. Kill output list and yank list to .faces file.
+;; 7. Execute `make test' in shell.
+
 ;;; php-mode-test.el ends here
